/*--------------------------------------------------------------------
 * $Id: rayleigh.c,v 1.2 2001/06/20 14:30:42 pa2h Exp $
 * 
 * This file is part of libRadtran.
 * Copyright (c) 1997-2001 by Arve Kylling and Bernhard Mayer.
 * 
 * ######### Contact info: http://www.libradtran.org #########
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License   
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.        
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the   
 * GNU General Public License for more details.                    
 * 
 * You should have received a copy of the GNU General Public License          
 * along with this program; if not, write to the Free Software                
 * Foundation, Inc., 59 Temple Place - Suite 330, 
 * Boston, MA 02111-1307, USA.
 *--------------------------------------------------------------------*/

#include <math.h>
#include "../src_f/f77-uscore.h"

/* @20i@                                                                  */
/* Function: rayleigh                                                     */
/* Description:                                                           */
/*  Calculates the Rayleigh scattering cross section as a function of     */
/*  wavelength. Uses the formula due to M. Nicolet, 'On the molecular     */
/*  scattering in the terrestrial atmosphere: An empirical formula for    */
/*  its calculation in the homosphere', Planet. Space Sci., vol. 32,      */
/*  pp. 1467-1468, 1984.                                                  */
/*                                                                        */
/* Parameters:                                                            */
/*    REAL(*) lambda: Wavelengths (nm) at which the Rayleigh scattering   */
/*                    cross section  is to be calculated. (input)         */
/*    INTEGER n_lambda: Number of wavelengths lambda. (input)             */
/*    INTEGER error:  Error flag. Not in use.                             */
/*    REAL(*) crs:    The Rayleigh cross section as a function of lambda. */
/*                    (output)                                            */
/* Return value:                                                          */
/* Example:                                                               */
/*  See @file{trans.f}.                                                   */
/* Files:                                                                 */
/*  @file{consts.int} must be present in the same directory               */
/*  as @file{rayleigh.f}.                                                 */
/* Known bugs:                                                            */
/* Author:  Arve Kylling                                                  */

/* @i20@ */

int F77_FCN (rayleigh) (float *crs, int *error, float *lambda, int *n_lambda)
{
  
  float x, lambda_um;
  int iv;
  float from_nm_um = 1./1000.;
  float ray_const = 4.02e-28;

  error = 0;
  for (iv = 0; iv <= *n_lambda; ++iv) {
    lambda_um = lambda[iv] * from_nm_um;
    if (lambda_um > .55) {
      x = .04;
    } else {
      x = lambda_um * .389 + .09426 / lambda_um - .3228;
    }
    crs[iv] = ray_const / pow(lambda_um, x+4.);
  }
  return 0;
}


