/*--------------------------------------------------------------------
 * $Id: aerosol.c,v 1.9 2001/07/11 09:48:37 pa2h Exp $
 * 
 * This file is part of libRadtran.
 * Copyright (c) 1997-2001 by Arve Kylling and Bernhard Mayer.
 * 
 * ######### Contact info: http://www.libradtran.org #########
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License   
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.        
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the   
 * GNU General Public License for more details.                    
 * 
 * You should have received a copy of the GNU General Public License          
 * along with this program; if not, write to the Free Software                
 * Foundation, Inc., 59 Temple Place - Suite 330, 
 * Boston, MA 02111-1307, USA.
 *--------------------------------------------------------------------*/

#include <stdio.h>							    
#include <stdlib.h>
#include <math.h>
#include "uvspec.h"
#include "../src_c/ascii.h"
#include "../src_c/miecalc.h"
#include "../src_c/fortran_and_c.h"
#include "../src_f/f77-uscore.h"

#define EPSILON 1E-6

/************************************/
/* prototypes of internal functions */
/************************************/

static int read_aerosol_profile (char *filename, int nlyr, int nlambda_t, 
				 float *zd, float **dtau);

static int read_aerosol_moments (char *filename,
				 float **moment, int *nmom);

static int scale_aer_tau (float **dtau, float *lambda_t, int nlambda_t, int nlyr,
			  float tau);

static int scale_aer_angstrom (float **dtau, float *lambda_t, int nlambda_t, int nlyr,
			       float alpha, float beta);

static int compare_levels (float *z1, int n1, float *z2, int n2);



/**************************************************************/
/* Read file holding  filenames for each layer.               */
/**************************************************************/

int read_optprop_files (char *filename, int nlyr, float *lambda_t, int nlambda_t, 
			 float *zd, float **dtau, float **ssa, float **gg,
			 float ***mom, int **nmom) 
{
  int i=0, iv=0, k=0, lc=0;
  int status=0;
  int rows=0, min_columns=0, max_columns=0, max_length=0;
  int arows=0, amin_columns=0, amax_columns=0;
  int n_tmp_lambda=0, tmp_nmom=0;
  int linear=1;
  float *tmp_zd=NULL, *tmp_lambda=NULL;
  float *tmp_dtau=NULL, *tmp_ssa=NULL, *tmp_gg=NULL;
  float **tmp_mom=NULL;
  float **data=NULL;
  float *temp=NULL, **temp_mom=NULL;
  char ***string=NULL;
  char *dummy=NULL;

  if ((status = ASCII_checkfile (filename, &rows, &min_columns, &max_columns, &max_length)) != 0) 
    return status;

  if (min_columns<2) {
    fprintf (stderr, "Error, need at least 2 columns in optical properties file %s\n", filename);
    return -1;
  }

  if ((status = ASCII_calloc_string(&string, rows, max_columns, FILENAME_MAX+1)) != 0) 
    return status;

  if ((status = ASCII_readfile(filename, string)) != 0) 
    return status;

  tmp_zd = calloc (rows, sizeof(double));
  
  for (i=0; i<rows; i++)
    tmp_zd[i] = (float) strtod (string[i][0], &dummy);

  
  /* check for consistency with atmospheric levels */
  status = compare_levels (tmp_zd, rows,
			   zd, nlyr+1);
  
  if (status!=0) {
    fprintf (stderr, "Error: z-grids of atmosphere and aerosols are different\n");
    return status;
  }



  temp = calloc (nlambda_t, sizeof(float));
					   
  /* For each layer read the optical properties file.                  */
  /* Note that the wavelength resolution may be different from the one */
  /* in the extraterrestrial solar flux file.                          */
  for (lc=0; lc<nlyr; lc++) {

    /* ??? here is potential for optimization! It is likely that    ??? */
    /* ??? files are used more than once, and if this is the case   ??? */
    /* ??? they should be copied from memory rather than read again ??? */

    if ((status = ASCII_file2float (string[lc+1][1], &arows, &amax_columns,
				    &amin_columns, &data)) != 0) {
      fprintf (stderr, "Error %d reading file %s for layer %d.\n", 
	       status, string[lc+1][1], lc);
      return status;
    }
    
    if (amin_columns<4) {
      fprintf (stderr, "Error, need at least 4 columns in optical properties file %s\n", 
	       string[lc+1][1]);
      return -1;
    }
    
    /* BBMM   tmp_nmom = amax_columns-3-1; */
    tmp_nmom = amax_columns-3;
    n_tmp_lambda = arows;
      
    tmp_lambda = calloc (n_tmp_lambda, sizeof(float));
    tmp_dtau   = calloc (n_tmp_lambda, sizeof(float));
    tmp_gg     = calloc (n_tmp_lambda, sizeof(float));
    tmp_ssa    = calloc (n_tmp_lambda, sizeof(float));
      
    /* BBMM if ((ASCII_calloc_float (&tmp_mom,  tmp_nmom+1, n_tmp_lambda)) != 0) */
    if ((ASCII_calloc_float (&tmp_mom,  tmp_nmom, n_tmp_lambda)) != 0)
      return status;

    for (iv=0;iv<n_tmp_lambda;iv++)
      tmp_lambda[iv] = data[iv][0];

    for (iv=0; iv<n_tmp_lambda; iv++) {
      tmp_dtau [iv] = data[iv][1];	 /* Still in units of km-1      */
      tmp_ssa  [iv] = data[iv][2];	
      /* BBMM for (k=0; k<=tmp_nmom; k++) */
      for (k=0; k<tmp_nmom; k++)
	if (data[iv][k+3]==data[iv][k+3]) /* Trick to replace NaN by 0  */
	  tmp_mom[k][iv] = data[iv][k+3];
	else
	  tmp_mom[k][iv] = 0;

      /* BBMM if (tmp_nmom>=1) */
      if (tmp_nmom>=2)
	tmp_gg [iv] = tmp_mom[1][iv];
      else 
	tmp_gg [iv] = 0;
    }

    ASCII_free_float (data, arows);

    for (iv=0; iv<n_tmp_lambda; iv++)
      tmp_dtau[iv] *= (tmp_zd[lc]-tmp_zd[lc+1]);
    
    /* Interpolate from optical property wavelengths to */
    /* the internal transmittance wavelength grid.      */
    
    /* interpolate optical thickness */
    status = arb_wvn(n_tmp_lambda, tmp_lambda, tmp_dtau,  
    		     nlambda_t, lambda_t, temp, linear);

    if (status!=0) {
      fprintf (stderr, "Error %d interpolating dtau in read_optprop_files()\n", status);
      return status;
    }

    for (iv=0; iv<nlambda_t; iv++)
      dtau[iv][lc] = temp[iv];

    /* interpolate asymmetry parameter */
    status = arb_wvn(n_tmp_lambda, tmp_lambda, tmp_gg,  
    		     nlambda_t, lambda_t, temp, linear);

    if (status!=0) {
      fprintf (stderr, "Error %d interpolating gg in read_optprop_files()\n", status);
      return status;
    }

    for (iv=0; iv<nlambda_t; iv++) 
      gg[iv][lc] = temp[iv];
    

    /* interpolate single scattering albedo */
    status = arb_wvn(n_tmp_lambda, tmp_lambda, tmp_ssa,
    		     nlambda_t, lambda_t, temp, linear);
    
    if (status!=0) {
      fprintf (stderr, "Error %d interpolating ssa in read_optprop_files()\n", status);
      return status;
    }
    
    for (iv=0; iv<nlambda_t; iv++) 
      ssa[iv][lc] = temp[iv];

    

    /* BBMM ASCII_calloc_float (&temp_mom, tmp_nmom+1, nlambda_t); */
    ASCII_calloc_float (&temp_mom, tmp_nmom, nlambda_t);

    /* BBMM for (k=0; k<=tmp_nmom; k++) { */
    for (k=0; k<tmp_nmom; k++) {
      status = arb_wvn(n_tmp_lambda, tmp_lambda, tmp_mom[k],
      		       nlambda_t, lambda_t, temp_mom[k], linear);
      if (status!=0) {
	fprintf (stderr, "Error %d interpolating mom in read_optprop_files()\n", status);
	return status;
      }
    }

    /* get index of maximum nonzero moment for each wavelength */
    for (iv=0; iv<nlambda_t; iv++) {
      /* BBMM for (k=tmp_nmom; k>=0; k--)   */
      for (k=tmp_nmom-1; k>=0; k--)  
	if (temp_mom[k][iv]!=0)
	  break;
      
      /* BBMM nmom[iv][lc] = k; */
      nmom[iv][lc] = k+1;
      
      /* BBMM mom[iv][lc] = calloc (nmom[iv][lc]+1, sizeof(float)); */
      mom[iv][lc] = calloc (nmom[iv][lc], sizeof(float));
      /* BBMM for (k=0; k<=nmom[iv][lc]; k++) */
      for (k=0; k<nmom[iv][lc]; k++)
	mom[iv][lc][k] = temp_mom[k][iv];
    }
      

    /* free memory */
    /* BBMM ASCII_free_float (temp_mom, tmp_nmom+1); */
    ASCII_free_float (temp_mom, tmp_nmom);

    /* BBMM ASCII_free_float (tmp_mom,  tmp_nmom+1); */
    ASCII_free_float (tmp_mom,  tmp_nmom);

    free (tmp_dtau);
    free (tmp_gg);
    free (tmp_ssa);
    free (tmp_lambda);
  }

  ASCII_free_string (string, rows, max_columns);
  free (tmp_zd);
  free (temp);
  
  return 0;
}



/**************************************************************/
/* Setup atmospheric aerosol.                                 */
/**************************************************************/

int setup_aerosol (input_struct input, output_struct *output)
{
  int status=0, ip=0, iv=0, lc=0, k=0, kmax=0;
  int maxawvn = MAXAWVN;

  int linear=0;
  int refrac=0, sizedist=0;

  float *tmp_aerext=NULL, *tmp_aerabs=NULL, *tmp_aersym=NULL, *tmp_aerwvn=NULL;  
  float *tmp_aer_dtau=NULL, *tmp_aer_gg=NULL, *tmp_aer_ssa=NULL;

  float *lref=NULL, *re=NULL, *im=NULL;
  int nref=0;

  double temperature = 300.0;  /* The temperature is only used for wavelengths > 167 um */
  double beta=0, beta0=0, ext=0, omega=0, g=0;
  
  mie_complex crefin;

  mie_inp_struct *mie_inp=NULL;
  mie_out_struct *mie_out=NULL;

  float *moment=NULL;
  int rows=0;
  

  int nmom=1000;   /* ??? should tie this to nstr to save computational time ??? */


  void F77_FCN (aerprof) (int* nlyr, float* zd, float* aer_pro, 
			  int* seasn, int* vulcan, float* visib );
  void F77_FCN (aeropt) ( int* vulcan, float* aerext, float* aerabs, 
			  float* aersym, float* aerwvn, float* rh, 
			  int* nlyr, float* zd, int* haze, int* maxawvn ); 
  void F77_FCN (aerwvn) ( float* aerwvn, float* aerabs, float* aersym, float* aerext,
			  float* aer_pro, float* aer_dtau, float* aer_gg, 
			  float* aer_ssa, int* nlyr, float* lambda_t, float* zd, 
			  int* nlambda_t, int* maxawvn );



  /* allocate memory for moments even if no aerosol specified */
  output->aer.nmom   = calloc (output->wl.nlambda_t, sizeof (int *));
  output->aer.moment = calloc (output->wl.nlambda_t, sizeof (float **)); 
  for (iv=0; iv<output->wl.nlambda_t; iv++) {
    output->aer.nmom  [iv] = calloc (output->atm.nlyr, sizeof (int));
    output->aer.moment[iv] = calloc (output->atm.nlyr, sizeof (float *));

    /* BBMM initialize nmom to -1; nmom==0 means an isotropic phase function */
    /* BBMM for (lc=0; lc<output->atm.nlyr; lc++) */
    /* BBMM output->aer.nmom [iv][lc] = -1; */
  }
  
  if (input.aer.vulcan > 0) {
    /**** Get appropriate aerosol profile and interpolate profile to zd grid ****/
    
    output->aer.aer_pro  = (float *) calloc(output->atm.nlev, sizeof(float));
    F77_FCN (aerprof) (&output->atm.nlyr, output->atm.zd, output->aer.aer_pro, 
		       &input.aer.seasn, &input.aer.vulcan, &input.aer.visib);
    
    /**** Get appropriate aerosol optical properties ****/
    
    tmp_aerext   = (float *) calloc (maxawvn*output->atm.nlev, sizeof(float));
    tmp_aerabs   = (float *) calloc (maxawvn*output->atm.nlev, sizeof(float));
    tmp_aersym   = (float *) calloc (maxawvn*output->atm.nlev, sizeof(float));
    tmp_aerwvn   = (float *) calloc (maxawvn*output->atm.nlev, sizeof(float));
    
    F77_FCN (aeropt) (&input.aer.vulcan, tmp_aerext, tmp_aerabs, 
		      tmp_aersym, tmp_aerwvn, output->atm.rh, 
		      &output->atm.nlyr, output->atm.zd, &input.aer.haze, &maxawvn); 
    
    tmp_aer_dtau = (float *) calloc (output->wl.nlambda_t*output->atm.nlyr, sizeof(float));
    tmp_aer_gg   = (float *) calloc (output->wl.nlambda_t*output->atm.nlyr, sizeof(float));
    tmp_aer_ssa  = (float *) calloc (output->wl.nlambda_t*output->atm.nlyr, sizeof(float));
    
    F77_FCN (aerwvn) (tmp_aerwvn, tmp_aerabs, tmp_aersym, tmp_aerext, output->aer.aer_pro, 
		      tmp_aer_dtau, tmp_aer_gg, tmp_aer_ssa, &output->atm.nlyr,
		      output->wl.lambda_t, output->atm.zd, &output->wl.nlambda_t, &maxawvn);
    
    output->aer.gg   = fortran2c_2D_float_ary (output->wl.nlambda_t, 
					       output->atm.nlyr, 
					       tmp_aer_gg );

    output->aer.ssa  = fortran2c_2D_float_ary (output->wl.nlambda_t, 
					       output->atm.nlyr, 
					       tmp_aer_ssa );

    output->aer.dtau = fortran2c_2D_float_ary (output->wl.nlambda_t, 
					       output->atm.nlyr, 
					       tmp_aer_dtau );
    
    free(tmp_aer_gg);
    free(tmp_aer_ssa);
    free(tmp_aer_dtau);
    
    free(tmp_aerext);
    free(tmp_aerabs);
    free(tmp_aersym);
    free(tmp_aerwvn);

    /**** Read file holding filenames of layer aerosol optical properties ****/
    /**** if specified                                                    ****/

    if (strlen(input.aer.filename[FN_AER_FILES]) > 0) {
      status = read_optprop_files (input.aer.filename[FN_AER_FILES], output->atm.nlyr,
				   output->wl.lambda_t, output->wl.nlambda_t, output->atm.zd,
				   output->aer.dtau, output->aer.ssa, output->aer.gg,
				   output->aer.moment, output->aer.nmom);
      if (status != 0) {
	fprintf (stderr, "Error %d reading aerosol optical property files: %s\n", 
		 status, input.aer.filename[FN_AER_FILES]);
	return status;
      }
      
      for (iv=0; iv<output->wl.nlambda_t; iv++)
	for (lc=0; lc<output->atm.nlyr; lc++)
	  /* BBMM if (output->aer.nmom[iv][lc] > output->atm.nmom) */
	  /* BBMM   output->atm.nmom = output->aer.nmom[iv][lc]; */
	  if (output->aer.nmom[iv][lc]-1 > output->atm.nmom)
	    output->atm.nmom = output->aer.nmom[iv][lc]-1;
    }

    
    /****   Read aerosol optical depth file if specified ****/
    if (strlen(input.aer.filename[FN_AER_TAU]) > 0) {
      status = read_aerosol_profile (input.aer.filename[FN_AER_TAU], output->atm.nlyr,
				     output->wl.nlambda_t, output->atm.zd,
				     output->aer.dtau);
      if (status != 0) {
	fprintf (stderr, "Error %d reading aerosol tau file: %s\n", 
		 status, input.aer.filename[FN_AER_TAU]);
	return status;
      }
    }


    /****   Read aerosol asymmetry parameter file if specified ****/
    if (strlen(input.aer.filename[FN_AER_GG]) > 0) {
      status = read_aerosol_profile (input.aer.filename[FN_AER_GG], output->atm.nlyr,
				     output->wl.nlambda_t, output->atm.zd,
				     output->aer.gg);
      if (status != 0) {
	fprintf (stderr, "Error %d reading aerosol asymmetry parameter file: %s\n", 
		 status, input.aer.filename[FN_AER_GG]);
	return status;
      }
    }


    /****   Read aerosol single scattering albedo file if specified ****/
    if (strlen(input.aer.filename[FN_AER_SSA]) > 0) {
      status = read_aerosol_profile (input.aer.filename[FN_AER_SSA], output->atm.nlyr,
				     output->wl.nlambda_t, output->atm.zd,
				     output->aer.ssa);
      if (status != 0) {
	fprintf (stderr, "Error %d reading aerosol single scattering albedo file: %s\n", 
		 status, input.aer.filename[FN_AER_SSA]);
	return status;
      }
    }
    
    
    /****   Read aerosol moments file if specified ****/
    /****   At present: independent of wavelength  ****/
    /****   but this can be easily changed!        ****/
    
    if (strlen(input.aer.filename[FN_AER_MOM]) > 0) {
      
      status = read_aerosol_moments (input.aer.filename[FN_AER_MOM], 
				     &moment, &rows);
      if (status != 0) {
	fprintf (stderr, "Error %d reading aerosol moments file: %s\n", 
		 status, input.aer.filename[FN_AER_MOM]);
	return status;
      }
      
      /* BBMM rows--; */

      /* copy to other wavelengths and layers */
      for (iv=0; iv<output->wl.nlambda_t; iv++)
	for (lc=0;lc<output->atm.nlyr;lc++) {
	  output->aer.nmom[iv][lc] = rows;
	  /* BBMM  output->aer.moment[iv][lc] = calloc (rows, sizeof (float)); */

	  output->aer.moment[iv][lc] = calloc (output->aer.nmom[iv][lc], sizeof (float));

	  /* BBMM  for (k=0; k<=output->aer.nmom[iv][lc]; k++) */
	  for (k=0; k<output->aer.nmom[iv][lc]; k++)
	    output->aer.moment[iv][lc][k] = moment[k];
	}

      /* increase number of moments if more found */
      for (iv=0; iv<output->wl.nlambda_t; iv++)
	for (lc=0; lc<output->atm.nlyr; lc++)
	  /* BBMM if (output->aer.nmom[iv][lc] > output->atm.nmom) */
	  /* BBMM   output->atm.nmom = output->aer.nmom[iv][lc]; */
	  if (output->aer.nmom[iv][lc]-1 > output->atm.nmom)
	    output->atm.nmom = output->aer.nmom[iv][lc]-1;
      

      free (moment);
    }

    /****   Read aerosol refractive index   ****/
    if (strlen(input.aer.filename[FN_AER_REF]) > 0) {

      /* read wavelength-dependent aerosol refractive index */
      status = read_3c_file_float (input.aer.filename[FN_AER_REF], 
				   &lref, &re, &im, &nref);
      
      if (status!=0) {
	fprintf (stderr, "Error %d reading aerosol refractive index file\n", status);
	return status;
      }

      /* linear interpolation of the refractive index to the model wavelengths */
      linear = 1;

      output->aer.re_t = (float *) calloc (output->wl.nlambda_t, sizeof(float));
      output->aer.im_t = (float *) calloc (output->wl.nlambda_t, sizeof(float));

      status = arb_wvn (nref, lref, re, 
			output->wl.nlambda_t, output->wl.lambda_t, output->aer.re_t, 
			linear);

      if (status != 0) {
	fprintf (stderr, "Error, status %d returned by arb_wvn()\n", status);
	return status;
      }

      status = arb_wvn (nref, lref, im, 
			output->wl.nlambda_t, output->wl.lambda_t, output->aer.im_t, linear);

      if (status != 0) {
	fprintf (stderr, "Error, status %d returned by arb_wvn()\n", status);
	return status;
      }
      
      free(lref); free(re); free(im);

      refrac=1;
    }
    else {
      if (input.aer.re>0) {   /* constant refractive index */
	output->aer.re_t = (float *) calloc (output->wl.nlambda_t, sizeof(float));
	output->aer.im_t = (float *) calloc (output->wl.nlambda_t, sizeof(float));

	for (iv=0; iv<output->wl.nlambda_t; iv++) { 
	  output->aer.re_t[iv] = input.aer.re;
	  output->aer.im_t[iv] = input.aer.im;
	}
	refrac=1;
      }
    }


    /****   Read aerosol size distribution  ****/
    if (strlen(input.aer.filename[FN_AER_SIZ]) > 0) {

      /* read size distribution */
      status = read_2c_file (input.aer.filename[FN_AER_SIZ], 
			     &(output->aer.xsiz), &(output->aer.ysiz), &(output->aer.nsiz));
      
      if (status!=0) {
	fprintf (stderr, "Error %d reading aerosol size distribution file\n", status);
	return status;
      }
      

      if (output->aer.xsiz[0]==0) {
	output->aer.xsiz[0] = 0.1 * output->aer.xsiz[1];
	if (!input.quiet)
	  fprintf (stderr, " ... aerosol size distribution: setting first data point to r=%g\n",
		   output->aer.xsiz[0]);
      }

      sizedist=1;
    }

    if (refrac&&sizedist) {

      if (!input.quiet)
	fprintf (stderr, " ... now doing aerosol Mie calculation\n");

      mie_inp = calloc (1, sizeof(mie_inp_struct));
      mie_out = calloc (1, sizeof(mie_out_struct));

      mie_out->pmom = calloc (4, sizeof(float *));
      for (ip=0; ip<4; ip++)
	mie_out->pmom[ip] = calloc (nmom+1, sizeof(float));
      

      /**********************************************************/
      /* The wavelength dependence of the optical thickness     */
      /* follows Mie theory; the absolute value of the optical  */
      /* thickness is defined by its value at this place of the */
      /* code; that is, the value defined by 'visibility' or    */
      /* (much better) by aerosol_set_tau. Later, it may still  */
      /* be scaled with the Angstrom parameters (if anybody     */
      /* wants that.						*/
      /* ??? I guess there are much better ways to do that ???  */
      /**********************************************************/

      for (iv=0; iv<output->wl.nlambda_t; iv++) {


	mie_inp->anyang      = 0;
	mie_inp->ipolzn      = 0;
	mie_inp->nmom        = nmom;
	mie_inp->momdim      = nmom+1;
	mie_inp->numang      = 0;
	mie_inp->mimcut      = 1.0e-08;
	mie_inp->perfct      = 0;
	mie_inp->prnt[0]     = 0;
	mie_inp->prnt[1]     = 0;
	mie_inp->xmu         = (float *)   calloc (mie_inp->numang+1, sizeof(float));
	mie_inp->s1          = (mie_complex *) calloc (mie_inp->numang+1, sizeof(mie_complex));
	mie_inp->s2          = (mie_complex *) calloc (mie_inp->numang+1, sizeof(mie_complex));


	crefin.re = output->aer.re_t[iv];
	crefin.im = output->aer.im_t[iv];

	status = mie_calc_sizedist (*mie_inp, mie_out, 
				    MIEV0, USER, crefin, 
				    (output->wl.lambda_t[iv])/1000.0, temperature,
				    output->aer.xsiz, output->aer.ysiz, output->aer.nsiz, 
				    &beta, &omega, &g);


	if (status!=0) {
	  fprintf (stderr, "Error %d returned by mie_calc_sizedist()\n", status);
	  return status;
	}

	/* determine maximum non-zero moment */
	kmax=0;
	for (k=1; k<=nmom; k++)
	  if (mie_out->pmom[0][k] != 0)
	    kmax=k;
	
	for (lc=0; lc<output->atm.nlyr; lc++)
	  output->aer.nmom[iv][lc] = kmax+1;

	/* copy moments to final destination */
	for (lc=0; lc<output->atm.nlyr; lc++) {
	  output->aer.moment[iv][lc] = calloc (output->aer.nmom[iv][lc], sizeof(float));
	  for (k=0; k<kmax; k++)
	    output->aer.moment[iv][lc][k] = mie_out->pmom[0][k]/mie_out->pmom[0][0];
	}
	

	/* extinction at first wavelength; required for scaling the optical thickness */
	if (iv==0)
	  beta0 = beta;


	ext = 0.0;
	  
	/* over-write optical thickness, single-scattering albedo, and asymmetry parameter */
	for (lc=0;lc<output->atm.nlyr;lc++) {
	  output->aer.dtau [iv][lc] *= beta/beta0;
	  output->aer.ssa  [iv][lc]  = omega;
	  output->aer.gg   [iv][lc]  = g;
	  
	  ext += output->aer.dtau[iv][lc];
	}
	

	if (!input.quiet) {
	  if (iv==0)
	    fprintf (stderr, "mie  lambda  nmom       tau       ssa       asy\n");
	  
	  fprintf (stderr, "mie %7.2f  %4d  %g  %g  %g\n",
		   output->wl.lambda_t[iv], kmax, ext, omega, g);
	}
      }

      /* free Mie memory */
      for (ip=0; ip<4; ip++)
	free (mie_out->pmom[ip]);
      free(mie_out->pmom);

      free (mie_inp);
      free (mie_out);
    }


    /**** Set aerosol asymmetry factor ****/
    if (input.aer.set_gg >= -1.0)
      for (lc=0;lc<output->atm.nlyr;lc++)
	for (iv=0;iv<output->wl.nlambda_t;iv++)
	  output->aer.gg[iv][lc] = input.aer.set_gg;
    
    /**** Set aerosol single scattering albedo ****/
    if (input.aer.set_ssa >= 0.0)
      for (lc=0;lc<output->atm.nlyr;lc++)
	for (iv=0;iv<output->wl.nlambda_t;iv++)
	  output->aer.ssa[iv][lc] = input.aer.set_ssa;
    
    /**** Scale aerosol single scattering albedo ****/
    if (input.aer.scale_ssa >= 0.0)
      for (lc=0;lc<output->atm.nlyr;lc++)
	for (iv=0;iv<output->wl.nlambda_t;iv++) {
	  output->aer.ssa[iv][lc] = output->aer.ssa[iv][lc] * input.aer.scale_ssa;
	  if (output->aer.ssa[iv][lc]>1.0) 
	    output->aer.ssa[iv][lc] = 1.0;
	}

    /**** Set aerosol optical thickness ****/
    if (input.aer.set_tau >= 0.0) {
      status = scale_aer_tau (output->aer.dtau, output->wl.lambda_t, output->wl.nlambda_t, 
			      output->atm.nlyr, input.aer.set_tau);

      if (status!=0) {
	fprintf (stderr, "Error %d returned by scale_aer_angstrom()\n", status);
	return status;

      }
    }

    /**** Scale aerosol optical thickness ****/
    if (input.aer.scale_tau >= 0.0) 
      for (lc=0;lc<output->atm.nlyr;lc++)
	for (iv=0;iv<output->wl.nlambda_t;iv++)
	  output->aer.dtau[iv][lc] = output->aer.dtau[iv][lc] * input.aer.scale_tau;


    /**** Set aerosol optical thickness according to Angstrom parameters ****/
    if (input.aer.beta >= 0.0) {
      status = scale_aer_angstrom (output->aer.dtau, output->wl.lambda_t, output->wl.nlambda_t, 
				   output->atm.nlyr, input.aer.alpha, input.aer.beta);

      if (status!=0) {
	fprintf (stderr, "Error %d returned by scale_aer_angstrom()\n", status);
	return status;

      }
    }
  }
  else {
    if ((status = ASCII_calloc_float(&output->aer.dtau, output->wl.nlambda_t, output->atm.nlyr)) != 0) return status;
    if ((status = ASCII_calloc_float(&output->aer.gg,   output->wl.nlambda_t, output->atm.nlyr)) != 0) return status;
    if ((status = ASCII_calloc_float(&output->aer.ssa,  output->wl.nlambda_t, output->atm.nlyr)) != 0) return status;
  }
  
  return 0;
}



/**************************************************************/
/* Read aerosol profile.                                      */
/**************************************************************/

static int read_aerosol_profile (char *filename, int nlyr, int nlambda_t, 
				 float *zd, float **dtau) 
{
  int rows=0;
  int i=0, j=0;
  int status = 0;
  float *aero_zd=NULL, *aero_tau=NULL;


  /* read altitude and optical depth from aerosol file */
  status = read_2c_file_float (filename, 
			       &aero_zd, &aero_tau, &rows);

    
  if (status!=0) {
    fprintf (stderr, "Error %d reading aerosol file %s\n", status, filename);
    return status;
  }

  /* check for consistency with atmospheric levels */
  status = compare_levels (aero_zd, rows,
			   zd, nlyr+1);
  
  if (status!=0) {
    fprintf (stderr, "Error: z-grids of atmosphere and aerosols are different\n");
    return status;
  }


  /* copy optical depth to dtau[][] */
  for (i=0;i<nlambda_t;i++)
    for (j=0;j<nlyr;j++)
      dtau[i][j] = aero_tau[j+1];

  /* free memory */
  free (aero_zd);
  free (aero_tau);

  return 0;
}



/**************************************************************/
/* Read moments of the aerosol phase function; memory for     */
/* float *moment is allocated automatically.                  */
/**************************************************************/

static int read_aerosol_moments (char *filename, 
	                         float **moment, int *nmom) 
{
  int i=0;
  int status = 0;
  double *aero_mom=NULL;
  
  /* read moments of the aerosol phase function from file */
  status = read_1c_file (filename, &aero_mom, nmom);
  
  if (status!=0) {
    switch(status)  {
    case ASCIIFILE_NOT_FOUND: 
      fprintf (stderr, "Aerosol tau file %s not found\n", filename);
      return status;
      break;
    case ASCII_NO_MEMORY: 
      fprintf (stderr, "Not enough memory for reading aerosol moments file %s\n",
	       filename);
      return status;
      break;
    default:
      fprintf (stderr, "Error %d reading aerosol file %s\n",  
	       status, filename);
      return status;
      break;
    }
  }


  /* normalize and copy moments to final destination */

  if (aero_mom[0]<=0) {
    fprintf (stderr, "Error, first moment cannot be zero or negative\n");
    return -1;
  }


  *moment = calloc (*nmom, sizeof(float));
  for (i=0;i<*nmom;i++)
    (*moment)[i] = aero_mom[i] / aero_mom[0];

  /* free memory */
  free (aero_mom);

  return 0;
}




/**************************************************************/
/* Scale the total aerosol optical thickness with Angstrom    */
/* parameters alpha and beta.                                 */
/**************************************************************/

static int scale_aer_angstrom (float **dtau, float *lambda_t, int nlambda_t, int nlyr,
			       float alpha, float beta)
{
  /* Local variables */
  float fact=0, depth=0;
  
  int lc=0, iv=0;
  float ext=0;
  
  
  /* Function Body */
  for (iv=0; iv<nlambda_t; iv++) {
    
    ext = 0.0;
    for (lc=0; lc<nlyr; lc++)
      ext += dtau[iv][lc];
    
    depth = beta * pow(lambda_t[iv]/1000.0, -alpha);
    fact = depth / ext;
    
    for (lc=0; lc<nlyr; lc++)
      dtau[iv][lc] *= fact;
  }
  
  return 0;
} 



/**************************************************************/
/* Scale the total aerosol optical thickness to a given value */
/**************************************************************/

static int scale_aer_tau (float **dtau, float *lambda_t, int nlambda_t, int nlyr,
			  float tau)
{
  /* Local variables */
  float fact=0;
  
  int lc=0, iv=0;
  float ext=0;
  
  
  /* Function Body */
  for (iv=0; iv<nlambda_t; iv++) {
    
    ext = 0.0;
    for (lc=0; lc<nlyr; lc++)
      ext += dtau[iv][lc];
    
    fact = tau / ext;
    
    for (lc=0; lc<nlyr; lc++)
      dtau[iv][lc] *= fact;
  }
  
  return 0;
} 



/*********************************/
/* compare two altitude grids    */
/*********************************/

static int compare_levels (float *z1, int n1,
			   float *z2, int n2)
{
  int kc=0;

  if (n1!=n2) {
    fprintf (stderr, "Error, number of layers differing\n");
    fprintf (stderr, "%d vs %d\n", n1, n2);

    return -1;
  }

  for (kc=0; kc<n1; kc++)
    if (fabs((float) (z1[kc]) - z2[kc]) > EPSILON * z1[kc]) {
      fprintf (stderr, "Error, difference at level %d, altitude %g vs. %g\n",	
	       kc, z1[kc], z2[kc]);
      return -1;
    }

  return 0;  /* if o.k. */
}


