/*--------------------------------------------------------------------
 * $Id: atmosphere.c,v 1.7 2001/07/11 09:48:37 pa2h Exp $
 * 
 * This file is part of libRadtran.
 * Copyright (c) 1997-2001 by Arve Kylling and Bernhard Mayer.
 * 
 * ######### Contact info: http://www.libradtran.org #########
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License   
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.        
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the   
 * GNU General Public License for more details.                    
 * 
 * You should have received a copy of the GNU General Public License          
 * along with this program; if not, write to the Free Software                
 * Foundation, Inc., 59 Temple Place - Suite 330, 
 * Boston, MA 02111-1307, USA.
 *--------------------------------------------------------------------*/

#include <stdio.h>							    
#include <stdlib.h>
#include "uvspec.h"
#include "../src_c/ascii.h"
#include "../src_c/radiative.h"
#include "../src_f/f77-uscore.h"



/************************************/
/* prototypes of internal functions */
/************************************/

static int read_atmosphere (char *filename, atm_inp_struct *atm_inp, 
			    atm_out_struct *atm_out, int other_column);

static int shift_atmosphere (atm_inp_struct *atm_inp, 
			     atm_out_struct *atm_out);
static int read_tau (char *filename, int nlyr,
		     float* zd, float** dtau);

static int read_molecular_absorption (char *filename, int nlyr,
				      float *zd, 
				      float **dtau_mono, 
				      float ***dtau_spec, float **lambda, int *nlambda,
				      int *monochromatic);


/**************************************************************/
/* Setup model atmosphere.                                    */
/**************************************************************/

int setup_atmosphere (input_struct input, output_struct *output)
{
  int i=0, status=0, lc=0, linear=0;
  float scale_factor=0;

  float *tmp_lambda=NULL, **tmp_dtau=NULL;
  int tmp_nlambda=0;

  float n_0=0, t_0=0, p_0=0, ar=0;

  int monochromatic=0;

  float F77_FCN (o3column) (float* dens, int* nlyr, float* zd);
  void  F77_FCN (scaleozone) (float* dens, int* nlyr, 
			      float* ozone_column, float* zd, float* ozone_scale_factor);


  /* read profiles */
  status = read_atmosphere(input.filename[FN_ATMOSPHERE], 
			   &input.atm, &output->atm, input.crs.column);

  if (status!=0) {
    fprintf (stderr, "Error %d reading atmosphere from %s\n",
	     status, input.filename[FN_ATMOSPHERE]);
    return status;
  }
  
  
  /* Shift atmosphere to given altitude */
  if (input.atm.altitude != 0) {
    status = shift_atmosphere(&input.atm, &output->atm);
    if (status !=0) {
      fprintf (stderr, "Error %d shifting atmosphere\n", status);
      return status;
    }
  }

  /* Scale profiles of pressure and air density */
  if (input.pressure >= 0.0) {
    scale_factor = input.pressure / output->atm.press[output->atm.nlyr];
    for (lc=0;lc<=output->atm.nlyr;lc++) {
      output->atm.press[lc] *= scale_factor;
      output->atm.dens[MOL_AIR][lc] *= scale_factor;
    } 
  }

  /* Scale ozone column */
  output->ozone_column = F77_FCN (o3column) (output->atm.dens[MOL_O3], &(output->atm.nlyr), output->atm.zd);
  
  if ( input.ozone_column >= 0.0 ) {
    F77_FCN (scaleozone) (output->atm.dens[MOL_O3], &output->atm.nlyr, &input.ozone_column,
			  output->atm.zd, &output->ozone_scale_factor);
    output->ozone_column = F77_FCN (o3column) (output->atm.dens[MOL_O3], &output->atm.nlyr, output->atm.zd); 
  }

  /* Calculate refractive index of air */
  if (input.rte.nrefrac > 0 ) {
    if (strlen(input.filename[FN_REFIND])>0) {
    }
    else {
      n_0 = 1.000293; /*         refraction under normal conditions */      
      t_0 = 15.0;     /*         normal temperature in celsius      */
      p_0 =1013.25;   /*         normal pression in mbar            */
      ar=0.00366;
      for (lc=0;lc<=output->atm.nlyr;lc++) {
	output->atm.refind[lc] = 1.0 + (n_0-1.0) *
	  (1.0+ar*t_0)/(1.0 + ar*(output->atm.temper[lc]-273.15)) *
	  output->atm.press[lc]/p_0;    
      }
    }
  } 
  else {
    for (lc=0;lc<=output->atm.nlyr;lc++) {
      output->atm.refind[lc] = 1.0;
    }    
  }


  /* Determine mixing ratios of trace gases and scale profiles */
  output->mixing_ratio = calloc (MX_NN, sizeof(float));

  /* CO2 */
  if (input.mixing_ratio[MX_CO2]<0) {
    /* If not specified, use the mixing ratio of the lowest level in the */
    /* atmospheric data file.                                            */

    output->mixing_ratio[MX_CO2] = output->atm.dens[MOL_CO2][output->atm.nlev-1] / 
      output->atm.dens[MOL_AIR][output->atm.nlev-1] * 1e6;
  }
  else {
    output->mixing_ratio[MX_CO2] = input.mixing_ratio[MX_CO2];
    
    for (i=0; i<output->atm.nlev; i++)
      output->atm.dens[MOL_CO2][i] = output->atm.dens[MOL_AIR][i] * input.mixing_ratio[MX_CO2] / 1e6;
  }
  

  /* H2O */
  if (input.mixing_ratio[MX_H2O]<0) {
    /* If not specified, use the mixing ratio of the lowest level in the */
    /* atmospheric data file.                                            */

    output->mixing_ratio[MX_H2O] = output->atm.dens[MOL_H2O][output->atm.nlev-1] / 
      output->atm.dens[MOL_AIR][output->atm.nlev-1] * 1e6;
  }
  else {
    output->mixing_ratio[MX_H2O] = input.mixing_ratio[MX_H2O];
    
    for (i=0; i<output->atm.nlev; i++)
      output->atm.dens[MOL_H2O][i] = output->atm.dens[MOL_AIR][i] * input.mixing_ratio[MX_H2O] / 1e6;
  }
  

  /* default CH4, N2O, F11, F12, and F22 are 'hardwired' */

  if (input.mixing_ratio[MX_CH4] < 0)
    output->mixing_ratio[MX_CH4] = 1.6; 
  else 
    output->mixing_ratio[MX_CH4] = input.mixing_ratio[MX_CH4];
  
  if (input.mixing_ratio[MX_N2O] < 0)
    output->mixing_ratio[MX_N2O] = 0.28; 
  else 
    output->mixing_ratio[MX_N2O] = input.mixing_ratio[MX_N2O];
  
  if (input.mixing_ratio[MX_F11] < 0)
    output->mixing_ratio[MX_F11] = 0.268E-3; 
  else 
    output->mixing_ratio[MX_F11] = input.mixing_ratio[MX_F11];
  
  if (input.mixing_ratio[MX_F12] < 0)
    output->mixing_ratio[MX_F12] = 0.503E-3; 
  else 
    output->mixing_ratio[MX_F12] = input.mixing_ratio[MX_F12];
  
  if (input.mixing_ratio[MX_F22] < 0)
    output->mixing_ratio[MX_F22] = 0.105E-3; 
  else 
    output->mixing_ratio[MX_F22] = input.mixing_ratio[MX_F22];
  

  /* Read Rayleigh cross section if specified */
  if (input.rayleigh == RAYLEIGH_FILE) {

    status = read_tau (input.filename[FN_RAYLEIGH], output->atm.nlyr,
		       output->atm.zd, &(output->atm.tau_rayleigh));
    if (status!=0) {
      fprintf (stderr, "Error, status %d returned by read_tau (\"%s\")\n", 
	       status, input.filename[FN_RAYLEIGH]);
      return status;
    }
  }			



  /* Read molecular absorption cross section if specified */
  if (strlen(input.filename[FN_MOLABS]) > 0) {
    status = read_molecular_absorption (input.filename[FN_MOLABS], output->atm.nlyr,
					output->atm.zd, 
					&(output->atm.tau_molabs), 
					&tmp_dtau, &tmp_lambda, &tmp_nlambda,
					&monochromatic);

    if (status!=0) {
      fprintf (stderr, "Error, status %d returned by read_molecular_absorption (\"%s\")\n", 
	       status, input.filename[FN_MOLABS]);
      return status;
    }

    if (monochromatic) {
      output->atm.molabs = MOLABS_FILE_MONO;

      if (!input.quiet)
	fprintf (stderr, " ... read monochromatic molecular absorption from %s\n", 
		 input.filename[FN_MOLABS]);
    }
    else {
      output->atm.molabs = MOLABS_FILE_SPEC;
      
      if (!input.quiet)
	fprintf (stderr, " ... read spectral molecular absorption from %s\n", 
		 input.filename[FN_MOLABS]);
      
      /* in the first version, we force the user to define his own transmittance grid */
      if (strlen(input.filename[FN_WLTRANS])==0) {
	fprintf (stderr, "Error. If a wavelength-dependent molecular absorption cross section\n");
	fprintf (stderr, "is specified in molecular_tau_file, a transmittance_wl_file needs\n");
	fprintf (stderr, "to be specified as well to define the computational wavelength grid;\n");
	fprintf (stderr, "e.g. for a line-by-line calculation, it is reasonable to use identical\n");
	fprintf (stderr, "grids for both molecular_tau_file and transmittance_wl_file.\n");

	return -1;
      }	

      
      /* interpolate optical thickness to transmittance grid */
      linear=1;

      /* allocate memory for optical thickness on internal transmittance grid */
      output->atm.tau_molabs_t = calloc (output->atm.nlyr, sizeof(float));
      for (lc=0; lc<output->atm.nlyr; lc++)
	output->atm.tau_molabs_t[lc] = calloc (output->wl.nlambda_t, sizeof(float));
      
      for (lc=0; lc<output->atm.nlyr; lc++) {
	status = arb_wvn (tmp_nlambda, tmp_lambda, tmp_dtau[lc], 
			  output->wl.nlambda_t, output->wl.lambda_t, output->atm.tau_molabs_t[lc],
			  linear);
	
	if (status != 0) {
	  fprintf (stderr, "Error %d interpolating molecular absorption grid to internal\n", status);
	  fprintf (stderr, "wavelength grid\n");
	  return status;
	}
      }


      free(tmp_lambda);
      for (lc=0; lc<output->atm.nlyr; lc++)
	free (tmp_dtau[lc]);

      free (tmp_dtau);
    }
  }			


  return 0;
}



/**************************************************************/
/* Setup temperature at output levels.                        */
/**************************************************************/

int setup_temperature (input_struct input, output_struct *output)
{
  int status=0, lev=0, iz=0;
  int linear=0;
  float *tmp_temper=NULL, *tmp_zd=NULL, *zout=NULL;

  /* Linearly interpolate output temperature to zout */
  /* sort profile in ascending order, assuming that */
  /* the original profile was descending            */
  
  tmp_zd     = (float *) calloc(output->atm.nlev, sizeof(float));
  tmp_temper = (float *) calloc(output->atm.nlev, sizeof(float));
  
  
  /* sort temperature profile in ascending order */
  for (lev=0; lev<output->atm.nlev; lev++) {
    tmp_zd[lev]     = output->atm.zd     [output->atm.nlev-1-lev];
    tmp_temper[lev] = output->atm.temper [output->atm.nlev-1-lev];
  }

  output->atm.temper_out  = (float *) calloc (input.atm.nzout, sizeof(float));
  linear = 1;

 
  /* Now we need to define a temporary zout grid;        */
  /* for rte_solver montecarlo it is possible to         */
  /* define an output altitude of -999 which stands      */
  /* for the surface altitude which may be a function    */
  /* of (x,y). This value is replaced by 0 which means   */
  /* that the surface temperature at the surface output  */
  /* level is set to temperature(z=0) which is not quite */
  /* correct if a 2D altitude grid has been defined.     */
  /* However, the only place where this information is   */
  /* currently used is the flexstor header which is      */
  /* required for photolysis calculations.               */

  zout = calloc (input.atm.nzout, sizeof(float));
  for (iz=0; iz<input.atm.nzout; iz++) {
    if (input.atm.zout[iz] != -999)
      zout[iz] = input.atm.zout[iz];
    else
      zout[iz] = 0;
  }
  
  /* interpolate temperature to zout grid */
  status = arb_wvn(output->atm.nlev, tmp_zd, tmp_temper, 
		   input.atm.nzout, zout, output->atm.temper_out, linear);

  if (status != 0) {
    fprintf (stderr, "Error interpolating temperature to output grid\n");
    return status;
  }


  free(tmp_temper); 
  free(tmp_zd);
  free(zout);


  /* Finally, copy the temperature of the lowest level to */
  /* output->temperature if the latter is not already set */

  if (output->surface_temperature < 0) {
    output->surface_temperature = output->atm.temper[output->atm.nlev-1];
    
    /* fprintf (stderr, " ... setting surface temperature to %f K\n", output->surface_temperature); */
  }

  return 0;
}




/**************************************************************/
/* Read atmosphere file and write to atm_out_struct           */
/**************************************************************/

static int read_atmosphere(char *filename, atm_inp_struct *atm_inp, 
			   atm_out_struct *atm_out, int other_column) 
{
  int rows=0, min_columns=0, max_columns=0;
  int status = 0;
  float **atm_data=NULL;
  
  status = ASCII_file2float (filename, 
			     &rows, &max_columns, &min_columns, 
			     &atm_data);

  if (status!=0) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }
  
  if (max_columns!=min_columns) {
    fprintf (stderr, "Warning: Inconsistent number of columns in %s\n", filename);
    fprintf (stderr, "     min = %d, max =%d\n", min_columns, max_columns);
    return -1;
  }  

  if (min_columns<2) {
    fprintf (stderr, "Error, too few columns in %s! Need at least %d.\n", filename, 2);
    return -1;
  }

  if (min_columns<other_column) {
    fprintf (stderr, "Error, too few columns in %s! Need at least %d.\n", filename, other_column);
    return -1;
  }
  
  atm_out->nlev = rows;
  atm_out->nlyr = rows-1;
  
  /* allocate memory for density array */
  atm_out->dens = calloc(atm_out->ndens, sizeof(float *));
  if (atm_out->dens == NULL) {
    fprintf (stderr, "Error allocating memory for atm_out->dens\n");
    return -1;
  }

  /* allocate memory for refractive index array */
  atm_out->refind = (float *) calloc(atm_out->nlyr+1, sizeof(float));
  if (atm_out->refind == NULL) {
    fprintf (stderr, "Error allocating memory for atm_out->refind\n");
    return -1;
  }

  atm_out->press     = ASCII_column_float(atm_data, atm_out->nlev, ATMFILE_PRESS );
  atm_out->temper    = ASCII_column_float(atm_data, atm_out->nlev, ATMFILE_TEMPER );
  atm_out->zd        = ASCII_column_float(atm_data, atm_out->nlev, ATMFILE_ZD );

  atm_out->dens[MOL_AIR] = ASCII_column_float(atm_data, atm_out->nlev, ATMFILE_AIR );
  atm_out->dens[MOL_O3]  = ASCII_column_float(atm_data, atm_out->nlev, ATMFILE_O3 );


  /* the following columns are optional */

  if (min_columns>5)
    atm_out->dens[MOL_O2]  = ASCII_column_float(atm_data, atm_out->nlev, ATMFILE_O2 );
  else 
    atm_out->dens[MOL_O2]  = (float *) calloc (atm_out->nlev, sizeof(float));


  if (min_columns>6)
    atm_out->dens[MOL_H2O] = ASCII_column_float(atm_data, atm_out->nlev, ATMFILE_H2O);
  else 
    atm_out->dens[MOL_H2O]  = (float *) calloc (atm_out->nlev, sizeof(float));
  

  if (min_columns>7)
    atm_out->dens[MOL_CO2] = ASCII_column_float(atm_data, atm_out->nlev, ATMFILE_CO2);
  else 
    atm_out->dens[MOL_CO2]  = (float *) calloc (atm_out->nlev, sizeof(float));


  if (min_columns>8)
    atm_out->dens[MOL_NO2] = ASCII_column_float(atm_data, atm_out->nlev, ATMFILE_NO2);
  else 
    atm_out->dens[MOL_NO2]  = (float *) calloc (atm_out->nlev, sizeof(float));


  atm_out->rh        = (float *) calloc (atm_out->nlev, sizeof(float));
  if (atm_out->rh == NULL) {
    fprintf (stderr, "Error allocating memory for atm_out->rh\n");
    return -1;
  }

  status = ASCII_free_float(atm_data, rows);
  return 0;
}




/**************************************************************/
/* Shift struct atm_out_struct to specified altitude.         */
/**************************************************************/

static int shift_atmosphere(atm_inp_struct *atm_inp, 
			    atm_out_struct *atm_out)
{
  int i=0, lev=0;

  float *z_shift = calloc(atm_out->nlev, sizeof(float));
  float press=0, temper=0, rh=0, air=0, o3=0, o2=0, h2o=0, co2=0, no2=0;
  
  /* calculate shifted levels (altitude now in km) */
  for (i=0; i<atm_out->nlev; i++)
    z_shift[i] = atm_out->zd[i] - atm_inp->altitude;


  /* check if the shifted profile extends down to 0 */
  if (z_shift[atm_out->nlev-1] > 0) {
    fprintf (stderr, "Error, the first level in the atmospheric data profile\n");
    fprintf (stderr, "is too high! It must be at least as low as ALTITUDE.\n");
    fprintf (stderr, "Found the first level at %f\n", 
	     atm_out->zd[atm_out->nlev-1]);
   return -1;
  }


  /* determine the first level below 0 */
  lev = atm_out->nlev;
  while (z_shift[--lev]<=0) {
    if (lev<=1) {
      fprintf (stderr, "Error, ALTITUDE is higher than the uppermost layer in the\n");
      fprintf (stderr, "atmospheric data profile.\n");
      return -1;
    }
  }

  /* lev is now the first level above 0; */
  /* interpolate profiles to new 0 level */

  press = atm_out->press[lev] + (0.0-z_shift[lev])*
    (atm_out->press[lev+1]-atm_out->press[lev])/(z_shift[lev+1]-z_shift[lev]);

  temper = atm_out->temper[lev] + (0.0-z_shift[lev])*
    (atm_out->temper[lev+1]-atm_out->temper[lev])/(z_shift[lev+1]-z_shift[lev]);

  rh = atm_out->rh[lev] + (0.0-z_shift[lev])*
    (atm_out->rh[lev+1]-atm_out->rh[lev])/(z_shift[lev+1]-z_shift[lev]);

  air = atm_out->dens[MOL_AIR][lev] + (0.0-z_shift[lev])*
    (atm_out->dens[MOL_AIR][lev+1]-atm_out->dens[MOL_AIR][lev])/(z_shift[lev+1]-z_shift[lev]);

  o3 = atm_out->dens[MOL_O3][lev] + (0.0-z_shift[lev])*
    (atm_out->dens[MOL_O3][lev+1]-atm_out->dens[MOL_O3][lev])/(z_shift[lev+1]-z_shift[lev]);

  o2 = atm_out->dens[MOL_O2][lev] + (0.0-z_shift[lev])*
    (atm_out->dens[MOL_O2][lev+1]-atm_out->dens[MOL_O2][lev])/(z_shift[lev+1]-z_shift[lev]);

  h2o = atm_out->dens[MOL_H2O][lev] + (0.0-z_shift[lev])*
    (atm_out->dens[MOL_H2O][lev+1]-atm_out->dens[MOL_H2O][lev])/(z_shift[lev+1]-z_shift[lev]);

  co2 = atm_out->dens[MOL_CO2][lev] + (0.0-z_shift[lev])*
    (atm_out->dens[MOL_CO2][lev+1]-atm_out->dens[MOL_CO2][lev])/(z_shift[lev+1]-z_shift[lev]);

  no2 = atm_out->dens[MOL_NO2][lev] + (0.0-z_shift[lev])*
    (atm_out->dens[MOL_NO2][lev+1]-atm_out->dens[MOL_NO2][lev])/(z_shift[lev+1]-z_shift[lev]);


  /* number of levels is probably smaller now */
  atm_out->nlev = lev+2;
  atm_out->nlyr = atm_out->nlev-1;

  z_shift[atm_out->nlev-1]=0;

  atm_out->press    [atm_out->nlev-1]  = press;
  atm_out->temper   [atm_out->nlev-1]  = temper;
  atm_out->rh       [atm_out->nlev-1]  = rh;

  atm_out->dens[MOL_AIR][atm_out->nlev-1]  = air;
  atm_out->dens[MOL_O3] [atm_out->nlev-1]  = o3;
  atm_out->dens[MOL_O2] [atm_out->nlev-1]  = o2;
  atm_out->dens[MOL_H2O][atm_out->nlev-1]  = h2o;
  atm_out->dens[MOL_CO2][atm_out->nlev-1]  = co2;
  atm_out->dens[MOL_NO2][atm_out->nlev-1]  = no2;

  for (lev=0; lev<atm_out->nlev; lev++)
    atm_out->zd[lev] = z_shift[lev];    


  return 0;  /* if o.k. */
}





/***********************************************************************************/
/*  Read a two-column file containing altitude [km] and optical thickness          */
/*  of each layer. Check if the levels equal those defined                         */
/*  in zd[]. Memory for dtau[] is allocated automatically.                         */ 
/***********************************************************************************/

static int read_tau (char* filename, int nlyr,
		     float* zd, float** dtau) 
{
  int rows=0;
  int lu=0;
  int status = 0;
  double *rayleigh_zd=NULL, *rayleigh_dtau=NULL;
  

  /* read altitude and optical depth from file */
  status = read_2c_file (filename, 
			 &rayleigh_zd, &rayleigh_dtau, &rows);

    
  if (status!=0) {
    switch(status)  {
    case ASCIIFILE_NOT_FOUND: 
      fprintf (stderr, "Error, %s not found\n", filename);
      return status;
      break;
    case ASCII_NO_MEMORY: 
      fprintf (stderr, "Error, not enough memory for reading %s\n",
	       filename);
      return status;
      break;
    case LESS_THAN_TWO_COLUMNS: 
      fprintf (stderr, "Error, %s has less than two columns\n",  
	       filename);
      return status;
      break;
    default:
      fprintf (stderr, "Error %d reading %s\n",  
	       status, filename);
      return status;
      break;
    }
  }

  /* check for consistency with atmospheric levels */

  if (rows != nlyr+1) {
    fprintf (stderr, "Error in %s: layering differs from atmospheric file:\n", filename); 
    fprintf (stderr, "atmosphere has %d levels, found %d in %s.\n", nlyr, rows, filename);
    status = 3;
    return status;
  }
  
  for (lu=0; lu<=nlyr; lu++) {
    if (zd[lu] != (float) rayleigh_zd[lu]) {
      fprintf (stderr, "Error, layers in %s differ from atmosphere layers: %d %f %f\n", 
	       filename, lu, rayleigh_zd[lu], zd[lu]);
      return -1;
    }
  }

  /* copy optical thickness to final destination */
  *dtau = calloc (nlyr, sizeof (float));
  for (lu=0; lu<nlyr; lu++)
    (*dtau)[lu] = rayleigh_dtau[lu+1];
    


  /* free memory */
  free (rayleigh_zd);
  free (rayleigh_dtau);

  return 0;
}




/***********************************************************************************/
/*  Read the molecular absorption file which can either be a two-column file       */
/*  containing altitude [km] and optical thickness of each layer or a matrix       */
/*  tau[wavelength, z]. Check if the levels equal those defined in zd[]. Memory    */
/*  for dtau[] is allocated automatically.                                         */ 
/***********************************************************************************/

static int read_molecular_absorption (char* filename, int nlyr,
				      float *zd, 
				      float **dtau_mono, 
				      float ***dtau_spec, float **lambda, int *nlambda,
				      int *monochromatic) 
{
  int rows=0, min_columns=0, max_columns=0, status=0;
  int lu=0, iv=0;


  float **data=NULL;
  

  /* read altitude and optical depth from file */
  status = ASCII_file2float (filename,   
			     &rows,        
			     &max_columns, 
			     &min_columns, 
			     &data);

  if (status!=0) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }


  if (min_columns<2) {
    fprintf (stderr, "Error, found less than 2 columns in %s\n", filename);
    return -1;
  }

  *monochromatic = 0;

  fprintf (stderr, "min_columns = %d, rows = %d, nlyr+1 = %d\n",
	   min_columns, rows, nlyr+1);
  fflush (stderr);

  if (min_columns==2 && rows == nlyr+1) {
    
    *monochromatic = 1;

    /* here we assume that the file comes in the old format,  */
    /* two columns z and dtau; in this case, the first column */
    /* must equal the atmospheric profile                     */
    
    fprintf (stderr, " ... monochromatic = ?\n");
    fflush (stderr);

    for (lu=0; lu<=nlyr; lu++) {
      if (zd[lu] != (float) data[lu][0])
	*monochromatic = 0;
    }

    fprintf (stderr, " ... monochromatic = %d\n", *monochromatic);
    fflush (stderr);

    if (!(*monochromatic)) {
      fprintf (stderr, "%s has two columns; nevertheless it is not interpreted as (z, dtau)\n", filename);
      fprintf (stderr, "format because values in the 1st column differ from atmospheric profile.\n");
    }
    else {

      /* copy optical thickness to final destination */
      *dtau_mono = calloc (nlyr, sizeof (float));
      for (lu=0; lu<nlyr; lu++)
	(*dtau_mono)[lu] = data[lu+1][1];
    }
  }
  
  
  if (!(*monochromatic)) {

    /* now we need to assume that we deal with a new format */
    /* where the first row must be the altitude levels      */
    
    fprintf (stderr, " ... now testing as wavelength-dependent absorption file\n");
    fflush (stderr);

    if (min_columns!=max_columns) {
      fprintf (stderr, "Error, inconsistent number of columns in %s, %d - %d\n",
	       filename, min_columns, max_columns);
      return -1;
    }


    if (min_columns<nlyr+1) {
      fprintf (stderr, "Error, expecting at least %d columns in %s\n", nlyr, filename);
      fprintf (stderr, "if it is supposed to be a wavelength-dependent file\n");

      return -1;
    }
    
    /* compare levels */
    for (lu=0; lu<=nlyr; lu++) {
      if (zd[lu] != (float) data[0][lu]) {
	fprintf (stderr, "Error, levels in %s differ from atmospheric levels: %f vs. %f\n", 
		 filename, zd[lu], data[0][lu]);
	return -1;
      }
    }


    /* now we know that the format is correct, and the only thing   */
    /* we need to do is to copy the data to their final destination */
    
    *nlambda = rows-1;
    *lambda = calloc (*nlambda, sizeof(float));
    *dtau_spec = calloc (nlyr, sizeof (float *));
    for (lu=0; lu<nlyr; lu++)
      (*dtau_spec)[lu] = calloc (*nlambda, sizeof(float));
    


    for (iv=0; iv<*nlambda; iv++) {
      (*lambda)[iv] = data[iv+1][0];
      
      for (lu=0; lu<nlyr; lu++)
	(*dtau_spec)[lu][iv] = data[iv+1][lu+1];
    }
      

  }

  return 0;
}

