/*--------------------------------------------------------------------
 * $Id: molecular.c,v 1.4 2001/06/29 13:51:25 pa2h Exp $
 * 
 * This file is part of libRadtran.
 * Copyright (c) 1997-2001 by Arve Kylling and Bernhard Mayer.
 * 
 * ######### Contact info: http://www.libradtran.org #########
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License   
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.        
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the   
 * GNU General Public License for more details.                    
 * 
 * You should have received a copy of the GNU General Public License          
 * along with this program; if not, write to the Free Software                
 * Foundation, Inc., 59 Temple Place - Suite 330, 
 * Boston, MA 02111-1307, USA.
 *--------------------------------------------------------------------*/

#include <stdio.h>							    
#include <stdlib.h>
#include <string.h>
#include "radiative.h"
#include "ascii.h"
#include "uvspec.h"

/************************************/
/* prototypes of internal functions */
/************************************/

static int crs_o3 (crs_out_struct *crs_out, int first, float *lambda, 
		   int n_lambda, int ntemp, char *path, float *temper,
		   int type);
static int closest_above(float lambda, float* lambda_raw, int n_crs); 
static int closest_below(float lambda, float* lambda_raw, int n_crs);



/**************************************************************/
/* Setup absorption by molecules                              */
/**************************************************************/

int setup_crs (input_struct input, output_struct *output)
{
  int status = 0;
  static int first = 1;
  
  /* Allocate memory for crs; currently Rayleigh scattering */
  /* and absorption by O3, O2, CO2, and H2O                 */

  if ( first ) {
    status  = ASCII_calloc_float_3D(&output->crs.crs, output->atm.nlev, 
				    MOL_NN, output->wl.nlambda_t );
    if (status!=0) { 
      fprintf (stderr, "Error %d allocating memory for output->crs.crs\n", status); 
      return status; 
    } 
  }

  switch(input.ck_scheme) {
  case CK_NONE:
    
    /* only ozone considered */
    status = crs_o3 (&output->crs,
		     first, output->wl.lambda_t, 
		     output->wl.nlambda_t, output->atm.nlev, input.filename[FN_PATH], 
		     output->atm.temper, input.crs.o3_type);
    if (status!=0) {
      fprintf (stderr, "Error %d setting up o3 cross section\n", status);
      return status;
    }

    break;

  case CK_FU:
  case CK_GENERIC:       /* ??? need to check if that is appropriate here ??? */
  case CK_AVHRR_KRATZ:   /* ??? need to check if this is appropriate here ??? */
    /* absorption coefficient profile */
    if (first) {
      status  = ASCII_calloc_float(&output->kabs.kabs, output->atm.nlev, 
				   output->wl.nlambda_t );
      if (status!=0) { 
	fprintf (stderr, "Error %d allocating memory for output->kabs.kabs\n", status); 
	return status; 
      } 
    }
    
    /* no break here, we need to call crs_ck */

  case CK_KATO:

    /* setup correlated k-tables */
    status = crs_ck (&(output->ck),
		     output->wl.lambda_t, output->wl.nlambda_t, 
		     output->atm.temper, output->atm.press, output->atm.zd,
		     output->atm.dens, output->atm.nlev, 
		     output->mixing_ratio, input.ck_h2ocont,
		     input.filename[FN_PATH], input.filename[FN_CK_GENERIC], first, &output->crs_ck);
		     
    if (status!=0) {
      fprintf (stderr, "Error %d setting up cross sections for correlated k\n", status);
      return status;
    }

    break;

  default:
    fprintf (stderr, "Error: unsupported correlated-k scheme\n");
    return -1;
    break;
  }


  /* Get additional requested cross sections */
  /**** FIXME: YET TO BE IMPLEMENTED   ****/
  if ( strlen(input.crs.filename) > 0 ) {
    printf("crs: %d %s\n",input.crs.column,input.crs.filename);
    return -1;
  }
  
  first=0;
  return 0;
}

/**************************************************************/
/* Setup molecular (Rayleigh) scattering                      */
/**************************************************************/

int setup_rayleigh (input_struct input, output_struct *output)
{
  int status=0;

  status = crs_rayleigh (output->wl.lambda_t, output->wl.nlambda_t, output->crs.crs);

  if (status!=0) {
    fprintf (stderr, "Error %d setting up Rayleigh cross section\n", status);
    return status;
  }

  return 0;
}



/*      SUBROUTINE crs_o3(crs, error, first, lambda, n_lambda, ntemp, 
 *     $     path, temper, type)  */
/*
*
*BOD
*DESCRIPTION
* Calculates ozone cross sections as a function of
* wavelength and temperature.
*USAGE
**FORTRAN
*     CALL crs_o3( crs, error, lambda, n_lambda, ntemp, path,
*    $             temper, type)  
**PERL
*INPUT
**first
*     logical
*      If true, temperature coefficients for ozone will be 
*      calculated at wavelengths lambda. If first is true, it will
*      be returned as false, just in case you should forget.
*      If false, the temperature coefficients from earlier call will be
*      used. 
*      first should be true the first time crs_o3 is called. It should 
*      also be true if lambda is changed. It does not need to be true
*      if temper is changed.
**lambda
*     real array(*)
*     Wavelengths (nm) at which the ozone cross section is to be calculated.
**n_lambda
*     integer
*     Number of wavelengths lambda.
**ntemp
*     integer
*     Number of different temperatures (levels in the atmosphere) 
**path
*     character string
*     The location of the Radtran data files, e.g. 
*     /usr/local/lib/Radtran/data
**temper
*     real array(*)
*     The temperatures (levels in the atmosphere) at which
*     the ozone cross section is to be calculated. In Kelvin.
**type
*     integer
*     The ozone cross section data to be used:
***1     Molina and Molina (1986)
***2     Daumont et al. (1992), Malicet et al. (1995)
***3     Paur and Bass (1985)
*OUTPUT
**crs
*     real array(temper,lambda)
*     The ozone cross section as a function of temper and lambda.
**error
*     integer
*     Error flag. Error checking performed only if first is true.
***0     Everything ok.
***1     Error occurred during reading of data file.
***2     n_lambda == 0 and ntemp != 0
***3     n_lambda != 0 and ntemp == 0
*ERROR_MESSAGES
*ROUTINES_CALLED
*FILES
*     crs_o3_mol_cf.dat           Reference 1 data file
*     crs_o3_dau_cf.dat           Reference 2 data file
*     crs_o3_pab_cf.dat           Reference 3 data file
*     consts.inc               Include file
*     param.inc                Include file
*REFERENCES
*     (1) Molina and Molina, 'Absolute Absorption Cross Sections of
*         Ozone in the 185- to 350-nm wavelength Range', JGR, vol 91,
*         pp. 14,501-14,508, 1986.
*
*     (2) Daumont D, Brion J, Charbonnier J and Malicet J, 
*         "Absorption Cross-Section at Room Temperature",
*         J.Atmos.Chem. no. 15, vol. 145 (1992)
*
*     (3) Paur and Bass, (1985) 'The ultraviolet cross sections of ozone, II, 
*         results and temperature dependence', in: Atmospheric Ozone, proceedings 
*         of the Quadrennial Ozone Symposium, Halkidiki, Greece. 
*         C. Zerefos and A. Ghazi (eds), pp.611-616, D. Reidel, Hingham, MA 
*
*EOD
*
*
*     References:
*     
*     WMO (1986), 'Atmospheric Ozone 1985', WMO Report No. 16.
*
*     */
int crs_o3 (crs_out_struct *crs_out, int first, float *lambda, 
	   int n_lambda, int ntemp, char *path, float *temper,
	   int type)
{
  int ia=0, ib=0, it=0, iv=0;
  int n_crs=0, status=0, max_columns=0, min_columns=0, rows=0;
  static int n_coeff=0;
  static float fact=0;
  char filenm[FILENAME_MAX+200]="", fullnm[FILENAME_MAX+200]="";
  static float *lambda_raw=NULL, **coeff_crs=NULL;
  float **data=NULL;
  float a=0.0, b=0.0, gradient=0.0;

  if ( first ) {
    if (n_lambda <= 0 && ntemp != 0)
      return 2;
    else { 
      if (n_lambda != 0 &&  ntemp <= 0)
	return 3;
    }

    switch (type) {
    case O3_MOLINA:
      strcpy(filenm,"crs/crs_o3_mol_cf.dat");
      n_coeff = 3;
      fact     = 1.E-20;
      break;
    case O3_DAUMONT:
      strcpy(filenm,"crs/crs_o3_dau_cf.dat");
      n_coeff = 3;
      fact     = 1.E-20;
      break;
    case O3_BASSPAUR:
      strcpy(filenm,"crs/crs_o3_pab_cf.dat");
      n_coeff = 3;
      fact     = 1.E-20;
      break;
    }
    
    strcpy(fullnm,path);
    strcat(fullnm,filenm);

    status = ASCII_file2float (fullnm, &rows, &max_columns, 
			       &min_columns, &data); 
    if (status!=0) { 
      fprintf (stderr, "Error %d opening file %s\n", status, fullnm); 
      return status; 
    } 

    n_crs  = rows;
    
    lambda_raw = (float *) calloc ((size_t) rows, sizeof (float));
    lambda_raw = ASCII_column_float (data, rows, 0);
    status  = ASCII_calloc_float(&coeff_crs, rows, max_columns);
    if (status!=0) { 
      fprintf (stderr, "Error %d allocating memory\n", status); 
      return status; 
    } 

    coeff_crs[0] =  ASCII_column_float (data, rows, 1);
    coeff_crs[1] =  ASCII_column_float (data, rows, 2);
    coeff_crs[2] =  ASCII_column_float (data, rows, 3);

    /* free memory */
    ASCII_free_float(data, rows);

  } /* if (first) */
  
  for (iv=0; iv<n_lambda; iv++) {
    ia=closest_above(lambda[iv], lambda_raw, n_crs);
    ib=closest_below(lambda[iv], lambda_raw, n_crs);

    if (ia > (n_crs-1) || ib<0) {
      fprintf (stderr, "* Error (crs_o3), wavelength %.1f nm out of range!\n", lambda[iv]);
      fprintf (stderr, "* Cannot do a monochromatic calculation at this wavelength;\n");
      fprintf (stderr, "* for infrared, have a look at the 'correlated_k' option!\n");
      fprintf (stderr, "* lambda: %f n_crs: %d ia: %d ib: %d\n",lambda[iv], n_crs, ia, ib);
      return 9;
    }

    for (it=0; it<ntemp; it++) {
      a = coeff_crs[0][ia] + 
	coeff_crs[1][ia] * (temper[it]-273.15) + 
	coeff_crs[2][ia]*(temper[it]-273.15)*(temper[it]-273.15);
      if (ia == ib) {
	crs_out->crs[it][MOL_O3][iv] = a*fact;
      }
      else {
	b=coeff_crs[0][ib] + 
	  coeff_crs[1][ib] * (temper[it]-273.15) + 
	  coeff_crs[2][ib]*(temper[it]-273.15)*(temper[it]-273.15);
	gradient=(a-b)/(lambda_raw[ia]-lambda_raw[ib]);
	crs_out->crs[it][MOL_O3][iv] = (gradient*(lambda[iv]-lambda_raw[ib])+b) * fact;
	/*printf("%d %d %f %e\n",it,iv,lambda[iv],crs_out->crs[it][MOL_O3][iv]);*/
      }
    }
  }

  return 0;
}

static int closest_above(float lambda, float* lambda_raw, int n_crs) 
{
  int iv;

  for (iv=n_crs-1;iv>=0;iv--)  {
    if ((lambda_raw[iv]-lambda) == 0.0) 
      return iv;
    else {
      if ((lambda_raw[iv]-lambda) < 0.0) 
	return iv+1;
    }
  }
  return 0;
}


static int closest_below(float lambda, float* lambda_raw, int n_crs) 
{
  int iv;
  for (iv=0; iv<n_crs; iv++) {
    if ( (lambda_raw[iv]-lambda) == 0.0) 
      return iv;
    else  {
      if ((lambda_raw[iv]-lambda) > 0.0) 
	return iv-1;
    }
  }
  return 0;
}
